from flask import Blueprint, request
from flask_jwt_extended import jwt_required
from flask_restful import Api, Resource

from portal.database.models import Course
from portal.rest import course_schema, courses_schema, course_import_schema
from portal.service import service
from portal.service.errors import PortalAPIError, ForbiddenError
from portal.tools.decorators import error_handler
from portal.tools.logging import log
from portal.service import policies

courses = Blueprint('courses', __name__)
courses_api = Api(courses)


class CourseResource(Resource):
    @error_handler
    @jwt_required
    def get(self, cid):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        # authorization
        if (policies.check_component(component=client, course=course, permissions=['view_course_full'])
                or policies.check_user(user=client, course=course, permissions=['view_course_full'])):
            return course_schema.dump(course)
        elif (policies.check_component(component=client, course=course, permissions=['view_course_limited'])
                or policies.check_user(user=client, course=course, permissions=['view_course_limited'])):
            return course_schema.dump(service.filter_course_info(course, client))

        raise ForbiddenError(uid=client.id)

    @error_handler
    @jwt_required
    def delete(self, cid):
        client = service.find_client()
        log.info(f"DELETE to {request.url}")
        course = service.find_course(cid)
        # authorization
        if policies.check_sysadmin(client):
            service.delete_entity(course)
            return '', 204

        raise ForbiddenError(uid=client.id)

    @error_handler
    @jwt_required
    def put(self, cid):
        log.info(f"PUT to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['update_course'])
                or policies.check_user(user=client, course=course, permissions=['update_course'])):
            raise ForbiddenError(uid=client.id)

        json_data = request.get_json()
        if not json_data:
            raise PortalAPIError(400, message='No data provided for course update.')

        data = course_schema.load(json_data)[0]

        course.name = data['name']
        course.codename = data['codename']
        service.write_entity(course)
        log.debug(f"Updated course {cid} to {course}.")
        return '', 204


class CourseList(Resource):
    @error_handler
    @jwt_required
    def get(self):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        # authorization
        if not policies.check_sysadmin(client):
            raise ForbiddenError(uid=client.id)

        courses = Course.query.all()
        return courses_schema.dump(courses)

    @error_handler
    @jwt_required
    def post(self):
        log.info(f"POST to {request.url}")
        client = service.find_client()
        # authorization
        if not policies.check_sysadmin(client):
            raise ForbiddenError(uid=client.id)

        json_data = request.get_json()
        if not json_data:
            raise PortalAPIError(400, message='No data provided for course creation.')

        data = course_schema.load(json_data)[0]

        name = data['name']
        codename = data['codename']
        new_course = Course(name=name, codename=codename)

        service.write_entity(new_course)
        log.debug(f"Created course={new_course}")
        return course_schema.dump(new_course)[0], 201


class CourseNotesToken(Resource):
    @error_handler
    @jwt_required
    def get(self, cid):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['handleNotesAccessToken'])
                or policies.check_user(user=client, course=course, permissions=['handleNotesAccessToken'])):
            raise ForbiddenError(uid=client.id)

        return course.notes_access_token

    @error_handler
    @jwt_required
    def put(self, cid):
        log.info(f"PUT to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['handleNotesAccessToken'])
                or policies.check_user(user=client, course=course, permissions=['handleNotesAccessToken'])):
            raise ForbiddenError(uid=client.id)

        json_data = request.get_json()
        if not json_data:
            raise PortalAPIError(400, message='No data provided for notes token update.')

        course.notes_access_token = json_data  # TODO: validation
        service.write_entity(course)
        log.info(f"Updated notes access token for course id={cid} to {json_data}")
        return '', 204


class CourseImport(Resource):
    @error_handler
    @jwt_required
    def put(self, cid):
        # TODO: import from IS MU
        log.info(f"PUT to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['update_course'])
                or policies.check_user(user=client, course=course, permissions=['update_course'])):
            raise ForbiddenError(uid=client.id)

        json_data = request.get_json()
        if not json_data:
            raise PortalAPIError(400, message='No data provided for notes token update.')
        data = course_import_schema.load(json_data)[0]
        source_course = service.find_course(data['source_course'])
        config = data['config']

        updated = service.copy_course(source_course, course, config)
        service.write_entity(updated)

        log.info(f"Updated course {cid} using data from {data['source_course']}.")
        return '', 204


courses_api.add_resource(CourseResource, '/<string:cid>')
courses_api.add_resource(CourseList, '')
courses_api.add_resource(CourseNotesToken, "/<string:cid>/notes_access_token")
courses_api.add_resource(CourseImport, "/<string:cid>/import")
