import json

from flask import Blueprint, request, jsonify
from flask_restful import Api, Resource, abort
from flask_jwt_extended import jwt_required

from portal.rest import project_schema, projects_schema, config_schema, submissions_schema, \
    submission_schema, submission_create_schema
from portal.tools.decorators import error_handler
from portal.tools.logging import log
from portal.database.models import Project, Submission, Course
from portal.service import service
from portal.service.errors import PortalAPIError, ForbiddenError
from portal.service import policies

projects = Blueprint('projects', __name__)
projects_api = Api(projects)

#update project state?
class ProjectResource(Resource):
    @error_handler
    @jwt_required
    def get(self, cid, pid):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        project = service.find_project(course, pid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['readProject'])
                or policies.check_user(user=client, course=course, permissions=['readProject'])):
            raise ForbiddenError(uid=client.id)

        return project_schema.dump(project)

    @error_handler
    @jwt_required
    def delete(self, cid, pid):
        log.info(f"DELETE to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        project = service.find_project(course, pid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['write_projects'])
                or policies.check_user(user=client, course=course, permissions=['write_projects'])):
            raise ForbiddenError(uid=client.id)

        service.delete_entity(project)
        return '', 204

    @error_handler
    @jwt_required
    def put(self, cid, pid):
        log.info(f"PUT to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        project = service.find_project(course, pid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['write_projects'])
                or policies.check_user(user=client, course=course, permissions=['write_projects'])):
            raise ForbiddenError(uid=client.id)

        json_data = request.get_json()
        if not json_data:
            raise PortalAPIError(400, message='No data provided for project update.')

        data = project_schema.load(json_data)[0]

        project.name = data['name']  # change to data.get (sysadmin)
        service.write_entity(project)
        log.debug(f"Updated project {project.name} to {project}.")
        return '', 204


class ProjectsList(Resource):
    @error_handler
    @jwt_required
    def get(self, cid):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['readProject'])
                or policies.check_user(user=client, course=course, permissions=['readProject'])):
            raise ForbiddenError(uid=client.id)
        return projects_schema.dump(course.projects)

    @error_handler
    @jwt_required
    def post(self, cid):
        log.info(f"POST to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['write_projects'])
                or policies.check_user(user=client, course=course, permissions=['write_projects'])):
            raise ForbiddenError(uid=client.id)

        json_data = request.get_json()
        if not json_data:
            raise PortalAPIError(400, message='No data provided for project creation.')

        data = project_schema.load(json_data)[0]

        new_project = Project(course=course, name=data['name'],
                              test_files_source=data['config']['test_files_source'])
        service.write_entity(new_project)
        log.debug(f"Created project={new_project} for course {cid}")
        return project_schema.dump(new_project)[0], 201


class ProjectConfigResource(Resource):
    @error_handler
    @jwt_required
    def get(self, cid, pid):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        project = service.find_project(course, pid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['readProject'])
                or policies.check_user(user=client, course=course, permissions=['readProject'])):
            raise ForbiddenError(uid=client.id)

        return config_schema.dump(project.config)

    @error_handler
    @jwt_required
    def put(self, cid, pid):
        log.info(f"PUT to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        project = service.find_project(course, pid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['write_projects'])
                or policies.check_user(user=client, course=course, permissions=['write_projects'])):
            raise ForbiddenError(uid=client.id)

        json_data = request.get_json()
        if not json_data:
            raise PortalAPIError(400, message='No data provided for project configuration update.')

        data = config_schema.load(json_data)[0]

        project.set_config(**data)
        service.write_entity(project)
        log.debug(f"Updated configuration for project {pid} in course {cid} to {data}.")
        return '', 204


class ProjectSubmissions(Resource):
    @error_handler
    @jwt_required
    def get(self, cid, pid):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        project = service.find_project(course, pid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['read_submissions_all'])
                or policies.check_user(user=client, course=course, permissions=['read_submissions_all'])):
            raise ForbiddenError(uid=client.id)

        user_id = request.args.get('user')
        if user_id:
            user = service.find_user(user_id)
            submissions = Submission.query.filter(Submission.user_id == user.id)\
                .filter(Submission.project_id == project.id)\
                .join(Course, Course.id == project.course_id).filter(Course.id == course.id)
            return submissions_schema.dump(submissions)
        return submissions_schema.dump(project.submissions)

    @error_handler
    @jwt_required
    def post(self, cid, pid):
        log.info(f"POST to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        project = service.find_project(course, pid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['create_submissions'])
                or policies.check_user(user=client, course=course, permissions=['create_submissions'])):
            raise ForbiddenError(uid=client.id)

        # check if a new submission can be created by this user in this project
        # TODO: configure conflict_timeout in project or remove completely (Kontr handles its own queue)
        if not service.can_create_submission(client, project):
            # maybe use 409, 403
            abort(429,
                  message=f"New submission for user {client.username} denied: a submission for this user "
                          f"is created or queued.")

        json_data = request.get_json()
        if not json_data:
            raise PortalAPIError(400, message='No data provided for submission creation.')

        data = submission_create_schema.load(json_data)[0]

        project_params = data.get(
            'project_params')  # data for Kontr processing, of no use to portal or submission q
        if project_params:
            project_params_string = json.dumps(project_params)  # not pretty, but should do for now
        else:
            project_params_string = ""
        new_submission = service.create_submission(user=client, project=project,
                                                   file_params=data['file_params'],
                                                   project_params_string=project_params_string)
        service.write_entity(new_submission)
        return submission_schema.dump(new_submission), 201


class ProjectTestFiles(Resource):
    @error_handler
    def get(self, cid, pid):
        # Gets test files (feature)
        return jsonify({"message": "Not implemented."})


projects_api.add_resource(ProjectResource, '/<string:pid>')
projects_api.add_resource(ProjectsList, '')
projects_api.add_resource(ProjectConfigResource, '/<string:pid>/config')
projects_api.add_resource(ProjectSubmissions, '/<string:pid>/submissions')
projects_api.add_resource(ProjectTestFiles, '/<string:pid>/files')
