from flask import Blueprint, request
from flask_jwt_extended import jwt_required
from flask_restful import Api, Resource

from portal.database.models import Role
from portal.rest import role_schema, roles_schema, users_schema, user_list_update_schema, permissions_schema
from portal.service import service
from portal.service.errors import PortalAPIError, ForbiddenError
from portal.tools.decorators import error_handler
from portal.tools.logging import log
from portal.service import policies

roles = Blueprint('roles', __name__)
roles_api = Api(roles)


class RoleResource(Resource):
    @error_handler
    @jwt_required
    def get(self, cid, rid):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        role = service.find_role(course, rid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['read_roles'])
                or policies.check_user(user=client, course=course, permissions=['read_roles'])):
            raise ForbiddenError(uid=client.id)

        return role_schema.dump(role)

    @error_handler
    @jwt_required
    def delete(self, cid, rid):
        log.info(f"DELETE to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        role = service.find_role(course, rid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['write_roles'])
                or policies.check_user(user=client, course=course, permissions=['write_roles'])):
            raise ForbiddenError(uid=client.id)
        service.delete_entity(role)
        return '', 204

    @error_handler
    @jwt_required
    def put(self, cid, rid):
        log.info(f"PUT to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        role = service.find_role(course, rid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['write_roles'])
                or policies.check_user(user=client, course=course, permissions=['write_roles'])):
            raise ForbiddenError(uid=client.id)

        json_data = request.get_json()
        if not json_data:
            raise PortalAPIError(400, message='No data provided for role update.')

        data = role_schema.load(json_data)[0]
        role.description = data['description']
        role.name = data['name']

        service.write_entity(role)
        log.debug(f"Updated role {rid} to {role}.")
        return '', 204


class RoleList(Resource):
    @error_handler
    @jwt_required
    def get(self, cid):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['read_roles'])
                or policies.check_user(user=client, course=course, permissions=['read_roles'])):
            raise ForbiddenError(uid=client.id)

        return roles_schema.dump(course.roles)

    @error_handler
    @jwt_required
    def post(self, cid):
        log.info(f"POST to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['write_roles'])
                or policies.check_user(user=client, course=course, permissions=['write_roles'])):
            raise ForbiddenError(uid=client.id)

        json_data = request.get_json()
        if not json_data:
            raise PortalAPIError(400, message='No data provided for role creation.')

        data = role_schema.load(json_data)[0]

        new_role = Role(course=course, name=data['name'])
        new_role.description = data.get('description') or ""
        service.write_entity(new_role)
        log.debug(f"Created role={new_role} for course {cid}")
        return role_schema.dump(new_role)[0], 201


class RolePermissions(Resource):
    @error_handler
    @jwt_required
    def get(self, cid, rid):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        role = service.find_role(course, rid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['read_roles'])
                or policies.check_user(user=client, course=course, permissions=['read_roles'])):
            raise ForbiddenError(uid=client.id)

        return permissions_schema.dump(role.permissions)

    @error_handler
    @jwt_required
    def put(self, cid, rid):
        log.info(f"PUT to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        role = service.find_role(course, rid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['write_roles'])
                or policies.check_user(user=client, course=course, permissions=['write_roles'])):
            raise ForbiddenError(uid=client.id)

        json_data = request.get_json()
        if not json_data:
            raise PortalAPIError(400, message='No data provided for update of role permissions.')

        data = permissions_schema.load(json_data)[0]

        role.set_permissions(**data)
        log.debug(f"Updated permissions for role {rid} in course {cid} to {data}.")
        return '', 204


class RoleUsersList(Resource):
    @error_handler
    @jwt_required
    def get(self, cid, rid):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        role = service.find_role(course, rid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['read_roles'])
                or policies.check_user(user=client, course=course, permissions=['read_roles'])):
            raise ForbiddenError(uid=client.id)

        return users_schema.dump(role.users)

    @error_handler
    @jwt_required
    def put(self, cid, rid):
        log.info(f"PUT to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        role = service.find_role(course, rid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['write_roles'])
                or policies.check_user(user=client, course=course, permissions=['write_roles'])):
            raise ForbiddenError(uid=client.id)

        json_data = request.get_json()

        data = user_list_update_schema.load(json_data)[0]

        # everything from users_add is added, THEN everything from users_remove is subtracted TODO: move to service
        users_old = set(role.users)
        users_add = data.get('add')
        users_remove = data.get('remove')
        users = set()
        if users_add:
            to_add = service.find_users(data['add'])
            users = users_old.union(to_add)
        if users_remove:
            to_remove = service.find_users(data['remove'])
            users = users_old.difference(to_remove)

        role.users = list(users)
        service.write_entity(role)
        log.info(f"Updated users of role {rid} to {data}.")
        return '', 204


class RoleUser(Resource):
    @error_handler
    @jwt_required
    def put(self, cid, rid, uid):
        log.info(f"PUT to {request.url}")
        client = service.find_client()
        log.debug(f"ADD User to role: course_id={cid}, role_id={rid}, user_id={uid}")
        course = service.find_course(cid)
        role = service.find_role(course, rid)
        user = service.find_user(uid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['assignRole'])
                or policies.check_user(user=client, course=course, permissions=['assignRole'])):
            raise ForbiddenError(uid=client.id)

        if user not in role.users:
            role.users.append(user)
            log.info(f"Added user {uid} to role {rid} in course {cid}.")
        else:
            log.info(f"User {uid} is already in role {rid} in course {cid}: no change.")
        return '', 204

    @error_handler
    @jwt_required
    def delete(self, cid, rid, uid):
        log.info(f"DELETE to {request.url}")
        client = service.find_client()
        log.debug(f"REMOVE User from role: course_id={cid}, role_id={rid}, user_id={uid}")
        course = service.find_course(cid)
        role = service.find_role(course, rid)
        user = service.find_user(uid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['assignRole'])
                or policies.check_user(user=client, course=course, permissions=['assignRole'])):
            raise ForbiddenError(uid=client.id)

        try:
            role.users.remove(user)
        except ValueError:
            raise PortalAPIError(400, message=f"Could not remove user {uid} from role {rid} in course {cid}: "
                                              f"role does not contain user.")
        log.info(f"Removed user {uid} from role {rid} in course {cid}.")
        return '', 204


roles_api.add_resource(RoleResource, "/<string:rid>")
roles_api.add_resource(RoleList, "")
roles_api.add_resource(RolePermissions, "/<string:rid>/permissions")
roles_api.add_resource(RoleUsersList, "/<string:rid>/users")
roles_api.add_resource(RoleUser, "/<string:rid>/users/<string:uid>")
